#include "launcher.h"

HINSTANCE hInst;
TCHAR g_szConnectAddress[128];

BOOL IsUnderWinMe(){
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&osvi);
	return (osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) && (osvi.dwMajorVersion <= 4) && (osvi.dwMinorVersion <= 90);
}

LPTSTR GetMyAddress(){
	static TCHAR h_name[256];
	struct hostent* host_ent;
	struct in_addr addr;

	gethostname(h_name, sizeof(h_name));

	host_ent = gethostbyname(h_name);

	if (host_ent->h_addr_list){
		int i = 0;
		while (host_ent->h_addr_list[i]){
			addr.S_un.S_addr = *(unsigned long*)host_ent->h_addr_list[i];
			if (htonl(addr.S_un.S_addr) >> 24 != 0x0A &&  //10.*.*.*
				htonl(addr.S_un.S_addr) >> 20 != 0xAC1 && //172.16-31.*.*
				htonl(addr.S_un.S_addr) >> 16 != 0xC0A8 && //192.168.*.*
				htonl(addr.S_un.S_addr) >> 16 != 0xA9FE)  //169.254.*.*
				return inet_ntoa(addr);
			i++;
		}
		addr.S_un.S_addr = *(unsigned long*)host_ent->h_addr_list[0];
		wsprintf(h_name, "%s (local)", inet_ntoa(addr));
		return h_name;
	}
	return "(unknown)";
}

LPTSTR GetPuyoInstallPath(){
#define PUYO_REG "Software\\Sega\\PuyoF2.0"

	HKEY hKey;
	DWORD dwcbData = MAX_PATH;
	static TCHAR s_szPuyoDir[MAX_PATH] = "";

	if (lstrlen(s_szPuyoDir) > 0)
		return s_szPuyoDir;

	if (RegOpenKeyEx(HKEY_CURRENT_USER, PUYO_REG, 0, 
		KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS){
		lstrcpy(s_szPuyoDir, "");
		return s_szPuyoDir;
	}
	
	if (RegQueryValueEx(hKey,"Install", NULL, NULL,
		(LPBYTE)s_szPuyoDir, &dwcbData) != ERROR_SUCCESS){
		lstrcpy(s_szPuyoDir, "");
	}

	RegCloseKey(hKey);

	return s_szPuyoDir;
#undef PUYO_REG
}

LPTSTR GetOldPuyoInstallPath(){
#define PUYO_REG "Software\\Sega\\PuyoF"

	HKEY hKey;
	DWORD dwcbData = MAX_PATH;
	static TCHAR s_szPuyoDir[MAX_PATH] = "";

	if (lstrlen(s_szPuyoDir) > 0)
		return s_szPuyoDir;

	if (RegOpenKeyEx(HKEY_CURRENT_USER, PUYO_REG, 0, 
		KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS){
		lstrcpy(s_szPuyoDir, "");
		return s_szPuyoDir;
	}
	
	if (RegQueryValueEx(hKey,"Install Path", NULL, NULL,
		(LPBYTE)s_szPuyoDir, &dwcbData) != ERROR_SUCCESS){
		lstrcpy(s_szPuyoDir, "");
	}

	RegCloseKey(hKey);

	return s_szPuyoDir;
#undef PUYO_REG
}

HRESULT App_SHGetFolderPath(HWND hwnd, int csidl, HANDLE hToken, DWORD dwFlags, LPTSTR pszPath){
	typedef HRESULT WINAPI PFUNC(HWND, int, HANDLE, DWORD, LPTSTR);
	typedef PFUNC FAR *__SHGetFolderPathA;
	HINSTANCE hFunc;
	HRESULT r_Result;
	__SHGetFolderPathA _SHGetFolderPath;

	if ((hFunc = LoadLibrary("Shell32.dll")) == NULL){ //Cũ[h
		return (HRESULT)0xFFFFFFFFL;
	}
	_SHGetFolderPath = (__SHGetFolderPathA)GetProcAddress(hFunc, "SHGetFolderPathA"); //AhX̎擾
	if (_SHGetFolderPath == NULL){ //s̎
		FreeLibrary(hFunc);
		return (HRESULT)0xFFFFFFFFL;
	}

	r_Result = _SHGetFolderPath(hwnd, csidl, hToken, dwFlags, pszPath);
	FreeLibrary(hFunc); //[hCůJ	
	return r_Result;
}

LPTSTR GetPuyoDataPath(){
	static TCHAR s_szAppDir[MAX_PATH] = "";
	HRESULT hRes;

	if (lstrlen(s_szAppDir) > 0)
		return s_szAppDir;

	hRes = App_SHGetFolderPath(NULL, CSIDL_APPDATA, NULL, 0, s_szAppDir);
	if (SUCCEEDED(hRes))
		lstrcat(s_szAppDir, "\\Sega\\PuyoF\\");
	else{
		lstrcpy(s_szAppDir, GetPuyoInstallPath());
	}

	return s_szAppDir;

}

void RunPuyoF(LPTSTR arg){
	TCHAR szPuyoPath[MAX_PATH];
	TCHAR szOption[512];
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	int mode_value;
	char mode_str[3];
	//char ping_str[5];

	memset(&si, 0, sizeof(STARTUPINFO));
	memset(&pi, 0, sizeof(PROCESS_INFORMATION));
	si.cb = sizeof(STARTUPINFO);

	lstrcpy(szPuyoPath, GetPuyoInstallPath());
	lstrcat(szPuyoPath, "\\PuyoF.exe");

	mode_value = 
		(GetAppSettingInt("s_launch", "Matchup_Mode", 1) & 3) << 4 |
		(GetAppSettingInt("s_launch", "Matchup_Rule", 0) & 1) << 3 |
		(GetAppSettingInt("s_launch", "Spectator_Mode", 0) & 3) << 1 |
		(GetAppSettingInt("s_launch", "Spectator_Rule", 0) & 1);
		if ((mode_value & 48) == 16) mode_value |= 32;
	itoa(mode_value, mode_str, 10);

	szOption[0] = 0;

	if (strchr(szPuyoPath, ' ') != NULL) lstrcat(szOption, "\"");
	lstrcat(szOption, szPuyoPath);
	if (strchr(szPuyoPath, ' ') != NULL) lstrcat(szOption, "\"");
	lstrcat(szOption, " ");
	lstrcat(szOption, arg);
	lstrcat(szOption, " ");
	lstrcat(szOption, GetAppSettingStr("s_launch", "Matchup_LoginID", ""));
	lstrcat(szOption, " ");
	lstrcat(szOption, GetAppSettingStr("s_launch", "Matchup_Password", ""));

	if (strcmp(arg, "S_MULTI1") == 0) {
		lstrcat(szOption, " ");
		lstrcat(szOption, mode_str);
		if ((mode_value & 48) == 32) {
			//itoa(Ping(), ping_str, 10);
			lstrcat(szOption, " ");
			lstrcat(szOption, GetAppSettingStr("s_launch", "Matchup_Ping", "5000"));
			//lstrcat(szOption, " ");
			//lstrcat(szOption, ping_str);
		} else if ((mode_value & 48) == 48) {
			lstrcat(szOption, " ");
			lstrcat(szOption, GetAppSettingStr("s_launch", "Matchup_Rate", "3000"));
		}
		if ((mode_value & 6) == 6) {
			lstrcat(szOption, " ");
			lstrcat(szOption, GetAppSettingStr("s_launch", "Spectator_Rate_Min", "2000"));
			lstrcat(szOption, " ");
			lstrcat(szOption, GetAppSettingStr("s_launch", "Spectator_Rate_Max", "4000"));
		}
	}

	if (!CreateProcess(szPuyoPath, szOption, NULL, NULL, FALSE, 0, NULL, GetPuyoInstallPath(), &si, &pi))
		MessageBox(NULL, "NɎs܂", APP_TITLE, MB_OK | MB_ICONHAND);

	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);
}

void RunPuyoF_old(LPTSTR arg){
	TCHAR szPuyoPath[MAX_PATH];
	TCHAR szOption[512];
	STARTUPINFO si;
	PROCESS_INFORMATION pi;

	memset(&si, 0, sizeof(STARTUPINFO));
	memset(&pi, 0, sizeof(PROCESS_INFORMATION));
	si.cb = sizeof(STARTUPINFO);

	lstrcpy(szPuyoPath, GetOldPuyoInstallPath());
	lstrcat(szPuyoPath, "PuyoF.exe");

	szOption[0] = 0;

	if (strchr(szPuyoPath, ' ') != NULL) lstrcat(szOption, "\"");
	lstrcat(szOption, szPuyoPath);
	if (strchr(szPuyoPath, ' ') != NULL) lstrcat(szOption, "\"");
	lstrcat(szOption, " ");
	lstrcat(szOption, arg);

	if (!CreateProcess(szPuyoPath, szOption, NULL, NULL, FALSE, 0, NULL, GetOldPuyoInstallPath(), &si, &pi))
		MessageBox(NULL, "NɎs܂", APP_TITLE, MB_OK | MB_ICONHAND);

	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);
}

void RunLobby(){
	TCHAR szLobbyPath[MAX_PATH];
	TCHAR szOption[512];
	STARTUPINFO si;
	PROCESS_INFORMATION pi;

	memset(&si, 0, sizeof(STARTUPINFO));
	memset(&pi, 0, sizeof(PROCESS_INFORMATION));
	si.cb = sizeof(STARTUPINFO);

	lstrcpy(szLobbyPath, GetPuyoInstallPath());
	lstrcat(szLobbyPath, "\\Lobby.exe");

	szOption[0] = 0;

	if (strchr(szLobbyPath, ' ') != NULL) lstrcat(szOption, "\"");
	lstrcat(szOption, szLobbyPath);
	if (strchr(szLobbyPath, ' ') != NULL) lstrcat(szOption, "\"");
	lstrcat(szOption, " ");
	lstrcat(szOption, "FromLauncher");
	lstrcat(szOption, " ");
	lstrcat(szOption, GetAppSettingStr("s_launch", "Matchup_LoginID", ""));
	lstrcat(szOption, " ");
	lstrcat(szOption, GetAppSettingStr("s_launch", "Matchup_Password", ""));

	if (!CreateProcess(szLobbyPath, szOption, NULL, NULL, FALSE, 0, NULL, GetPuyoInstallPath(), &si, &pi))
		MessageBox(NULL, "NɎs܂", APP_TITLE, MB_OK | MB_ICONHAND);

	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);
}

void LoadAddrHistory(HWND hctrl){
	INT i;
	TCHAR szSec[16];
	PCHAR p;

	for (i = 0;i < 16;i++){
		wsprintf(szSec, "Addr%02d", i+1);
		p = GetAppSettingStr("Recent Connection", szSec, "");
		if (lstrlen(p) == 0)
			break;
		ComboBox_AddString(hctrl, p);
	}
}

void SpaceCut(LPTSTR psz){
	// Space R[h΋IɏI[
	PCHAR p;
	p = psz;
	while (*p){
		if (IsDBCSLeadByte(*p))
			p++;
		else if (*p == ' '){
			*p = 0;
			break;
		}
		p++;
	}
}

void SaveAddrHistory(HWND hctrl, LPCTSTR szNewCmd){
	INT i, j, len;
	TCHAR szSec[16];
	PCHAR pszData;

	if (lstrlen(szNewCmd) == 0)
		return;

	//܂VACeǉ
	SetAppSettingStr("Recent Connection", "Addr01", szNewCmd);
	j = 1;
	for (i = 0;j < 16;i++){
		len = ComboBox_GetLBTextLen(hctrl, i);
		if (len == CB_ERR)
			break;
		pszData = GlobalAlloc(GPTR, len + 1);
		ComboBox_GetLBText(hctrl, i, pszData);
		if (lstrcmpi(szNewCmd, pszData) == 0)
			continue;
		wsprintf(szSec, "Addr%02d", ++j);
		SetAppSettingStr("Recent Connection", szSec, pszData);
		GlobalFree(pszData);
	}
}

int MakeRTrayMenu(HWND hWnd)
{
	HMENU hMenu, hSubMenu;
	POINT pt;

	hMenu = LoadMenu(hInst, MAKEINTRESOURCE(IDM_TRAYMENU));
	hSubMenu = GetSubMenu(hMenu, 0);
	GetCursorPos(&pt);
	SetForegroundWindow(hWnd);
	TrackPopupMenu(
		hSubMenu, TPM_BOTTOMALIGN, pt.x, pt.y, 0, hWnd, NULL);
	DestroyMenu(hMenu);
	return 0;
}

INT_PTR CALLBACK ConnectDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam){
	switch (uMsg){
	case WM_INITDIALOG:
		LoadAddrHistory(GetDlgItem(hDlg, IDC_CON_ADDR));
		SetFocus(GetDlgItem(hDlg, IDC_CON_ADDR));
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam)){
		case IDOK:
			GetDlgItemText(hDlg, IDC_CON_ADDR, g_szConnectAddress, 128);
			SpaceCut(g_szConnectAddress);
			SaveAddrHistory(GetDlgItem(hDlg, IDC_CON_ADDR), g_szConnectAddress);
			EndDialog(hDlg, 1);
			break;
		case IDCANCEL:
			EndDialog(hDlg, 0);
			break;
		}
		break;
	case WM_CLOSE:
		EndDialog(hDlg, 0);
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

INT_PTR CALLBACK MainDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam){
	static HICON hIcon;
	static NOTIFYICONDATA nid;
	static BOOL isIcon = FALSE;
	int ping_ms;
	char ping_str[16];

	switch (uMsg){
	case WM_INITDIALOG:
		SetWindowText(hDlg, APP_TITLE);
		hIcon = LoadIcon(hInst, MAKEINTRESOURCE(IDI_ICONMAIN));
		SendMessage(hDlg, WM_SETICON, ICON_SMALL, (LPARAM)hIcon);
		SetDlgItemText(hDlg, IDC_IPADDR, GetMyAddress());
		ping_ms = Ping();
		_snprintf(ping_str, 16, "%d ~b", ping_ms);
		SetDlgItemText(hDlg, IDC_PING, ping_str);

		nid.cbSize = sizeof(NOTIFYICONDATA);
		nid.hWnd = hDlg;
		nid.uID = IDT_TRAY;
		nid.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
		nid.uCallbackMessage = WM_USER;
		nid.hIcon = hIcon;
		strcpy(nid.szTip, APP_TITLE);
		break;
	case WM_APP:
		if (isIcon) {
			isIcon = FALSE;
			Shell_NotifyIcon(NIM_DELETE, &nid);
		}

		ShowWindow(hDlg, SW_SHOWNORMAL);
		SetForegroundWindow(hDlg);
		break;
	case WM_USER:
		if (wParam == IDT_TRAY){
			switch (lParam){
			case WM_LBUTTONDOWN:
				isIcon = FALSE;
				Shell_NotifyIcon(NIM_DELETE, &nid);
				ShowWindow(hDlg, SW_SHOWNORMAL);
				break;
			case WM_RBUTTONDOWN:
				MakeRTrayMenu(hDlg);
				break;
			}
			break;
		}
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam)){
		case IDCANCEL:
			if (MessageBox(NULL, "IꍇAՂՂtB[o[ƃr[I܂B\n{ɏI܂H", APP_TITLE, MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2 | MB_SETFOREGROUND) == IDYES){
				if (isIcon){
					isIcon = FALSE;
					Shell_NotifyIcon(NIM_DELETE, &nid);
				}
				EndDialog(hDlg, 0);
			}
			break;
		case IDC_B_OFFLINE:
			RunPuyoF("SINGLE");
			if (!isIcon) {
				isIcon = TRUE;
				Shell_NotifyIcon(NIM_ADD, &nid);
				ShowWindow(hDlg, SW_HIDE);
			}
			break;
		case IDC_B_MATCHUP:
			RunPuyoF("S_MULTI1");
			if (!isIcon) {
				isIcon = TRUE;
				Shell_NotifyIcon(NIM_ADD, &nid);
				ShowWindow(hDlg, SW_HIDE);
			}
			break;
		case IDC_B_LOBBY:
			RunLobby();
			if (!isIcon) {
				isIcon = TRUE;
				Shell_NotifyIcon(NIM_ADD, &nid);
				ShowWindow(hDlg, SW_HIDE);
			}
			break;
		case IDC_B_HOST:
			RunPuyoF_old("D_HOST");
			if (!isIcon) {
				isIcon = TRUE;
				Shell_NotifyIcon(NIM_ADD, &nid);
				ShowWindow(hDlg, SW_HIDE);
			}
			break;
		case IDC_B_JOIN:
			if (DialogBox(hInst, MAKEINTRESOURCE(IDD_CONNECT), hDlg, ConnectDlgProc)){
				TCHAR szCmd[256];
				wsprintf(szCmd, "D_JOIN %s", g_szConnectAddress);
				RunPuyoF_old(szCmd);
				if (!isIcon) {
					isIcon = TRUE;
					Shell_NotifyIcon(NIM_ADD, &nid);
					ShowWindow(hDlg, SW_HIDE);
				}
			}
			break;
		case IDC_B_CONFIG:
			DoSetup(hDlg, 0);
			break;
		}
		break;
	case WM_SYSCOMMAND:
		switch (wParam){
		case SC_MINIMIZE:
			isIcon = TRUE;
			Shell_NotifyIcon(NIM_ADD, &nid);
			ShowWindow(hDlg, SW_HIDE);
			return TRUE;
		default: 
			return FALSE;
		}
	break;
	case WM_DESTROY:
		DestroyIcon(hIcon);
		break;
	case WM_CLOSE:
		if (MessageBox(NULL, "IꍇAՂՂtB[o[ƃr[I܂B\n{ɏI܂H", APP_TITLE, MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2 | MB_SETFOREGROUND) == IDYES){
			if (isIcon){
				isIcon = FALSE;
				Shell_NotifyIcon(NIM_DELETE, &nid);
			}
			EndDialog(hDlg, 0);
		}
		break;
	default:
		return FALSE;
	}
	return TRUE;
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow){
	WSADATA wd;
	HANDLE mut;
	HWND hDlg;

//	INITCOMMONCONTROLSEX icx;
	if (IsUnderWinMe()){
		MessageBox(NULL, "Win9xɂ͑ΉĂ܂B", APP_TITLE, MB_OK | MB_ICONHAND);
		return 0;
	}

	if (WSAStartup(MAKEWORD(1, 1), &wd)){
		MessageBox(NULL, "\Pbg̏Ɏs", APP_TITLE, MB_OK | MB_ICONHAND);
		return 0;
	}

	hDlg = FindWindow(NULL, APP_TITLE);
	if (hDlg) {
		PostMessage(hDlg, WM_APP, 0, 0);
		return 2;
	}

	hInst = hInstance;

	mut = CreateMutex(NULL, TRUE, "Global\\PUYOSTARTMUTEX");

	InitCommonControls();
//	icx.dwSize = sizeof(INITCOMMONCONTROLSEX);
//	icx.dwICC = ICC_WIN95_CLASSES;
//	InitCommonControlsEx();

	DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAINDLG), NULL, MainDlgProc);

	if (mut != NULL) CloseHandle(mut);

	WSACleanup();
	return 1;
}
